use crate::StdErr;
use futures::stream::StreamExt;
use rand::seq::SliceRandom;
use rand::{prelude::ThreadRng, thread_rng};
use std::cell::RefCell;
use std::sync::Arc;
use tracing::{event, instrument, Level};
use twilight_gateway::{Cluster, Event};
use twilight_http::Client;
use twilight_model::{
    channel::message::Message,
    gateway::Intents,
    id::{marker::UserMarker, Id},
};

thread_local!(static RNG: RefCell<ThreadRng> = RefCell::new(thread_rng()));

const GREETINGS: &'static [&'static str] =
    &["You rung?", "Hello there", "Right back atcha", "Yes?"];

pub async fn run_discord(token: &str) -> StdErr<()> {
    event!(Level::INFO, "Starting...");
    let (cluster, mut events) = Cluster::new(
        token.to_owned(),
        Intents::GUILD_MESSAGES | Intents::MESSAGE_CONTENT,
    )
    .await?;
    tokio::spawn(async move {
        cluster.up().await;
    });
    let client = Arc::new(Client::new(token.to_owned()));
    while let Some((shard_id, ev)) = events.next().await {
        let client = client.clone();
        tokio::spawn(async move {
            if let Err(e) = handle_event(ev, shard_id, client).await {
                event!(Level::ERROR, err=?e, "Error handling discord event");
            }
        });
    }
    Ok(())
}

fn mentions(msg: &Message, us: Id<UserMarker>) -> bool {
    msg.mentions.iter().filter(|m| m.id == us).next().is_some()
}

#[instrument(skip(client))]
async fn handle_event(ev: Event, shard_id: u64, client: Arc<Client>) -> StdErr<()> {
    let me = client.current_user().exec().await?.model().await?;
    match ev {
        Event::MessageCreate(msg) => {
            if mentions(&msg.0, me.id) {
                let greet = RNG.with(|rng| GREETINGS.choose(&mut *rng.borrow_mut()).unwrap());
                client
                    .create_message(msg.channel_id)
                    .content(greet)?
                    .exec()
                    .await?;
            }
        }
        Event::ShardConnected(_) => event!(Level::INFO, "Shard connected"),
        _ => {}
    }
    Ok(())
}
